#!/usr/bin/python
# -*- coding: utf-8 -*-
"""ORM for index server."""
from time import sleep
from datetime import datetime
from peewee import Model, OperationalError, BooleanField
from peewee import CharField, IntegerField, BigIntegerField, TextField, DateTimeField, DecimalField
from playhouse.migrate import SchemaMigrator, migrate
from playhouse.db_url import connect
from pacifica.ingest.config import get_config

SCHEMA_MAJOR = 2
SCHEMA_MINOR = 0
DB = connect(get_config().get('database', 'peewee_url'))


class OrmSync(object):
    """
    Special module for syncing the orm.

    This module should incorporate a schema migration strategy.

    The supported versions migrating forward must be in a versions array
    containing tuples for major and minor versions.

    The version tuples are directly translated to method names in the
    orm_update class for the update between those versions.

    Example Version Control::

      class orm_update:
        versions = [
          (0, 1),
          (0, 2),
          (1, 0),
          (1, 1)
        ]

        def update_0_1_to_0_2():
            pass
        def update_0_2_to_1_0():
            pass

    The body of the update should follow peewee migration practices.
    http://docs.peewee-orm.com/en/latest/peewee/playhouse.html#migrate
    """

    versions = [
        (0, 0),
        (1, 0),
        (2, 0)
    ]

    @staticmethod
    def dbconn_blocking():
        """Wait for the db connection."""
        dbcon_attempts = get_config().getint('database', 'connect_attempts')
        dbcon_wait = get_config().getint('database', 'connect_wait')
        while dbcon_attempts:
            try:
                IngestState.database_connect()
                return
            except OperationalError:
                # couldnt connect, potentially wait and try again
                sleep(dbcon_wait)
                dbcon_attempts -= 1
        raise OperationalError('Failed database connect retry.')

    @classmethod
    def update_0_0_to_1_0(cls):
        """Update by creating the table."""
        if not IngestState.table_exists():
            IngestState.create_table()
        col_names = [col_md.name for col_md in DB.get_columns('ingeststate')]
        if 'complete' in col_names:
            migrator = SchemaMigrator(DB)
            migrate(migrator.drop_column(
                'ingeststate', 'complete'
            ))

    @classmethod
    def update_1_0_to_2_0(cls):
        """Update by adding the boolean column."""
        migrator = SchemaMigrator(DB)
        migrate(migrator.add_column(
            'ingeststate', 'complete',
            BooleanField(default=False)
        ))

    @classmethod
    def update_tables(cls):
        """Update the database to the current version."""
        verlist = cls.versions
        db_ver = IngestStateSystem.get_version()
        if verlist.index(verlist[-1]) == verlist.index(db_ver):
            # we have the current version don't update
            return
        with IngestState.atomic():
            for db_ver in verlist[verlist.index(db_ver):-1]:
                next_db_ver = verlist[verlist.index(db_ver)+1]
                method_name = 'update_{}_to_{}'.format(
                    '{}_{}'.format(*db_ver),
                    '{}_{}'.format(*next_db_ver)
                )
                getattr(cls, method_name)()
            IngestStateSystem.drop_table()
            IngestStateSystem.create_table()
            IngestStateSystem.get_or_create_version()


# pylint: disable=too-few-public-methods
class BaseModel(Model):
    """Auto-generated by pwiz."""

    class Meta(object):
        """Map to the database connected above."""

        database = DB


class IngestStateSystem(BaseModel):
    """Ingest State Schema Version Model."""

    part = CharField(primary_key=True)
    value = IntegerField(default=-1)

    @classmethod
    def get_or_create_version(cls):
        """Set or create the current version of the schema."""
        if not cls.table_exists():
            return (0, 0)
        major, _created = cls.get_or_create(part='major', value=SCHEMA_MAJOR)
        minor, _created = cls.get_or_create(part='minor', value=SCHEMA_MINOR)
        return (major, minor)

    @classmethod
    def get_version(cls):
        """Get the current version as a tuple."""
        if not cls.table_exists():
            return (0, 0)
        return (cls.get(part='major').value, cls.get(part='minor').value)

    @classmethod
    def is_equal(cls):
        """Check to see if schema version matches code version."""
        major, minor = cls.get_version()
        return major == SCHEMA_MAJOR and minor == SCHEMA_MINOR

    @classmethod
    def is_safe(cls):
        """Check to see if the schema version is safe for the code."""
        major, _minor = cls.get_version()
        return major == SCHEMA_MAJOR


class IngestState(BaseModel):
    """Map a python record to a mysql table."""

    job_id = BigIntegerField(primary_key=True, column_name='id')
    state = CharField()
    task = CharField()
    task_percent = DecimalField()
    exception = TextField(default='')
    complete = BooleanField(default=False)
    created = DateTimeField(default=datetime.utcnow)
    updated = DateTimeField(default=datetime.utcnow)

    @classmethod
    def atomic(cls):
        """Get the atomic context or decorator."""
        # pylint: disable=no-member
        return cls._meta.database.atomic()
        # pylint: enable=no-member

    @classmethod
    def database_connect(cls):
        """
        Make sure database is connected.

        Trying to connect a second
        time *does* cause problems.
        """
        # pylint: disable=no-member
        if not cls._meta.database.is_closed():
            cls._meta.database.close()
        cls._meta.database.connect()
        # pylint: enable=no-member

    @classmethod
    def database_close(cls):
        """
        Close the database connection.

        Closing already closed database
        is not a problem, so continue on.
        """
        # pylint: disable=no-member
        if not cls._meta.database.is_closed():
            cls._meta.database.close()
        # pylint: enable=no-member

    class Meta(object):
        """Map to uniqueindex table."""

        table_name = 'ingeststate'
# pylint: enable=too-few-public-methods


def update_state(job_id, state, task, task_percent, exception=''):
    """Update the state of an ingest job."""
    completed = False
    if state == 'FAILED' or (task == 'ingest metadata' and task_percent == 100):
        completed = True
    if job_id and int(job_id) >= 0:
        IngestState.database_connect()
        record = IngestState.get_or_create(job_id=job_id,
                                           defaults={'task': '', 'task_percent': 0, 'state': ''})[0]

        record.state = state
        record.task = task
        record.task_percent = task_percent
        record.exception = exception
        record.complete = completed
        record.updated = datetime.utcnow()
        record.save()
        IngestState.database_close()


def read_state(job_id):
    """Return the state of an ingest job as a json object."""
    IngestState.database_connect()
    if job_id and job_id >= 0:
        record = IngestState.get(job_id=job_id)
    else:
        record = IngestState()
        record.state = 'DATA_ACCESS_ERROR'
        record.task = 'read_state'
        record.task_percent = 0
    IngestState.database_close()
    return record



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Testable utilities for ingest."""
from __future__ import print_function
import json
import requests
import six
from .config import get_config

# pylint: disable=invalid-name
int_type = six.integer_types[-1]
# pylint: enable=invalid-name


def parse_size(size):
    """Parse size string to integer."""
    units = {
        'B': 1, 'KB': 10**3, 'MB': 10**6, 'GB': 10**9, 'TB': 10**12,
        'b': 1, 'Kb': 1024, 'Mb': 1024**2, 'Gb': 1024**3, 'Tb': 1024**4
    }
    number, unit = [string.strip() for string in size.split()]
    return int_type(float(number)*units[unit])


def create_state_response(record):
    """Create the state response body from a record."""
    return {
        'job_id': record.job_id,
        'state': record.state,
        'task': record.task,
        'task_percent': str(record.task_percent),
        'complete': bool(record.complete),
        'updated': str(record.updated),
        'created': str(record.created),
        'exception': str(record.exception)
    }


def get_unique_id(id_range, mode):
    """Return a unique job id from the id server."""
    uniqueid_url = get_config().get('uniqueid', 'url')
    url = '{0}/getid?range={1}&mode={2}'.format(
        uniqueid_url, id_range, mode)

    req = requests.get(url)
    body = req.text
    info = json.loads(body)
    unique_id = info['startIndex']

    return unique_id



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Test ingest with good uploads of good and bad data."""
from __future__ import print_function, absolute_import
import os
import json
import hashlib
from contextlib import contextmanager
from tempfile import NamedTemporaryFile
try:
    from tempfile import TemporaryDirectory
except ImportError:  # pragma: no cover
    from backports.tempfile import TemporaryDirectory
from six import ensure_binary
import requests
from common_methods_test import try_good_move, check_upload_state, try_assert_job_state


def test_good_move():
    """Test the good move."""
    try_good_move('move-md', 'OK', 'ingest metadata', 100)


def test_bad_file_move():
    """Test the bad file path move."""
    try_good_move('bad-move-md', 'FAILED', 'move files', 0)


@contextmanager
def create_temp_files(temp_dir, num_files=100):
    """Create lots of temporary files and save them."""
    file_list = []
    for file_index in range(num_files):
        file_name = ''
        file_hash = ''
        with NamedTemporaryFile(dir=temp_dir, suffix='.txt', delete=False) as temp_file:
            temp_file.write(ensure_binary('This is a temp file {}.'.format(file_index)))
            file_name = temp_file.name
        with open(file_name) as file_desc:
            file_hash = hashlib.sha1()
            file_hash.update(ensure_binary(file_desc.read()))
            file_hash = file_hash.hexdigest()
        file_list.append({
            'ctime': 'Tue Nov 29 14:09:05 PST 2016',
            'destinationTable': 'Files',
            'hashsum': file_hash,
            'hashtype': 'sha1',
            'mimetype': 'text/plain',
            'mtime': 'Tue Nov 29 14:09:05 PST 2016',
            'name': 'tempfile.{}.txt'.format(file_index),
            'size': os.stat(file_name).st_size,
            'source': file_name,
            'subdir': 'data/a/b/'
        })
    yield file_list


def test_big_move():
    """Test should create a large move and verify speed."""
    with open(os.path.join('test_data', 'move-md.json'), 'r') as md_fd:
        md_data = json.loads(md_fd.read())
        md_data.pop()
    with TemporaryDirectory() as temp_dir:
        with create_temp_files(temp_dir) as file_list:
            md_data.extend(file_list)
            req = requests.post(
                'http://127.0.0.1:8066/move',
                json=md_data,
                headers={'content-type': 'application/json'}
            )
            assert req.status_code == 200
            job_id = req.json()['job_id']
            job_state = check_upload_state(job_id, 10)
            try_assert_job_state(job_state, 'OK', 'ingest metadata', 100)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Configuration reading and validation module."""
from os import getenv
try:
    from ConfigParser import SafeConfigParser
except ImportError:  # pragma: no cover python 2 vs 3 issue
    from configparser import ConfigParser as SafeConfigParser
from pacifica.ingest.globals import CONFIG_FILE


def get_config():
    """Return the ConfigParser object with defaults set."""
    configparser = SafeConfigParser()
    configparser.add_section('ingest')
    configparser.set('ingest', 'transfer_size', getenv(
        'TRANSFER_SIZE', '4 Mb'))
    configparser.set('ingest', 'volume_path', getenv(
        'VOLUME_PATH', '/tmp'))
    configparser.add_section('uniqueid')
    configparser.set('uniqueid', 'url', getenv(
        'UNIQUEID_URL', 'http://127.0.0.1:8051'))
    configparser.add_section('policy')
    configparser.set('policy', 'ingest_url', getenv(
        'POLICY_INGEST_URL', 'http://127.0.0.1:8181/ingest'))
    configparser.add_section('archiveinterface')
    configparser.set('archiveinterface', 'url', getenv(
        'ARCHIVEINTERFACE_URL', 'http://127.0.0.1:8080'))
    configparser.add_section('metadata')
    configparser.set('metadata', 'ingest_url', getenv(
        'METADATA_INGEST_URL', 'http://127.0.0.1:8121/ingest'))
    configparser.add_section('database')
    configparser.set('database', 'peewee_url', getenv(
        'PEEWEE_URL', 'sqliteext:///db.sqlite3'))
    configparser.set('database', 'connect_attempts', getenv(
        'DATABASE_CONNECT_ATTEMPTS', '10'))
    configparser.set('database', 'connect_wait', getenv(
        'DATABASE_CONNECT_WAIT', '20'))
    configparser.add_section('celery')
    configparser.set('celery', 'broker_url', getenv(
        'BROKER_URL', 'pyamqp://'))
    configparser.set('celery', 'backend_url', getenv(
        'BACKEND_URL', 'rpc://'))
    configparser.read(CONFIG_FILE)
    return configparser



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Test ingest."""
import os
import mock
import peewee
import requests
from pacifica.ingest.orm import OrmSync, IngestState


@mock.patch.object(IngestState, 'database_connect')
def test_bad_db_connection(mock_is_table_exists):
    """Test a failed db connection."""
    mock_is_table_exists.side_effect = peewee.OperationalError(
        mock.Mock(), 'Error')
    hit_exception = False
    os.environ['DATABASE_CONNECT_ATTEMPTS'] = '1'
    os.environ['DATABASE_CONNECT_WAIT'] = '1'
    try:
        OrmSync.dbconn_blocking()
    except peewee.OperationalError:
        hit_exception = True
    assert hit_exception


def test_rest_root_status():
    """Test the root level status page."""
    resp = requests.get('http://127.0.0.1:8066')
    assert resp.status_code == 200, 'Status code should be 200 OK'
    assert 'message' in resp.json(), 'Status should be object with message key.'
    assert resp.json()['message'] == 'Pacifica Ingest Up and Running', 'message should be specific.'
